import pandas as pd
import numpy as np

from pandas.api.types import is_datetime64_dtype, is_float_dtype, is_int64_dtype


class Processor:
    """
    Main class the pre-process time-series aggregated wearable data. The available methods of the class should be
    used in a chaining style. It also offers a "magic" method "process" that processes the data in a pre-defined
    suggested pipeline oriented for physical activity data.

    Attributes:
        df (pd.DataFrame): The dataframe holding the uni-variate time-series data.
    """

    def __init__(self, df):
        self.__check_correctness(df)
        self.df = df

    def process(self, granularity, q, impute_start=None, impute_end=None):
        """
        A "magic" all-in-one method that holds a proposed logic for a series of suggested pre-processing steps for
        aggregated wearable data.

        Args:
            granularity (str): The string in panda's offset style for resampling and aggregating the time series data.
            q (float): Value between 0 <= q <= 1, the quantile(s) to compute for outliers removal.
            impute_start (int): The starting hour in 24-hours format, for imputation of zeros.
            impute_end (int): The ending hour in 24-hours format, for imputation of zeros.

        Returns:
            (pd.DataFrame): The pre-processed dataset.
        """

        self\
            .remove_duplicate_values_at_same_timestamp()\
            .remove_nan() \
            .remove_outlier_values(q=q) \
            .resample_dates(frequency=granularity)

        # Imputation is only meaningful in hourly granularity
        if granularity == '1H':
            self.impute_zeros(start_hour=impute_start, end_hour=impute_end)

        self.add_date_features() \
            .add_sin_cos_features(keep_only_sin_cos_transforms=True) \

        processed_df = self.df
        return processed_df

    def remove_nan(self):
        """
        Removes NaT/NaN dates that break the resampling.

        Returns:
            (self).
        """

        # NaN dates are also represented as NaT (Not A Timestamp)
        self.df.dropna(inplace=True)
        return self

    def remove_outlier_values(self, q=0.05):
        """
        Removes outlier steps count data by using the quantile method.

        Args:
           q (float): Value between 0 <= q <= 1, the quantile(s) to compute

        Returns:
            (self).
        """

        q_low = self.df["value"].quantile(q)
        q_hi = self.df["value"].quantile(1 - q)
        self.df = self.df[(self.df["value"] < q_hi) & (self.df["value"] > q_low)]
        return self

    def resample_dates(self, frequency):
        """
        Resamples and aggregates the data in fixed intervals and sums the data.

        Args:
            frequency (str): The string in panda's offset style for resampling and aggregating the time series data

        Returns:
            (self).
        """
        # check if after outliers removal, df is empty then return the empty df
        # e.g. (the same object as is)
        if self.df.empty:
            return self

        self.df = self.df.resample(rule=frequency).sum()
        return self

    def impute_zeros(self, start_hour=8, end_hour=24):
        """
        Imputes zero steps values in specific times using the interpolation method.
        By default it imputes zeros between 08:00 to 24:00. If there are data between this time period, and
        zeros too, then interpolates those zeros.
        The intuition is not to impute every zero since the zero values in the midnight and first morning hours are
        realistic.
        This function has meaning and applies only when resampling by hour e.g. for hourly datasets and not for daily
        datasets.

        Args:
            start_hour (int): The starting hour in 24-hours format, for imputation of zeros.
            end_hour (int): The ending hour in 24-hours format, for imputation of zeros.

        Returns:
            (self).

        """

        mask = (self.df.index.hour >= start_hour) & (self.df.index.hour <= end_hour) & (self.df['value'] == 0)
        # Replace those zeros with NaN to be caught by the pandas' interpolate method.
        self.df.loc[mask] = np.nan
        self.df.interpolate(method='linear', inplace=True, limit_direction='forward')
        return self

    def remove_duplicate_values_at_same_timestamp(self):
        """
        Removes the exactly identical records based on the time (startTime) and the according value of steps.

        Returns:
            (self).
        """

        # Drop subsequent days that have
        # at the exact same timestamp with the exact same value of steps
        self.df.drop_duplicates(subset=['value'], keep='first', inplace=True)
        return self

    def add_date_features(self):
        """
        Adds date features to the user's data.

        Returns:
            (self).
        """

        self.df['dayofweek'] = self.df.index.dayofweek
        self.df['week'] = self.df.index.isocalendar().week
        self.df['month'] = self.df.index.month
        self.df['year'] = self.df.index.year
        self.df['day'] = self.df.index.day
        self.df['hour'] = self.df.index.hour

        return self

    def add_sin_cos_features(self, keep_only_sin_cos_transforms=True):
        """
        Adds sin/cos transformation as features from the plain date features, to encode the cyclical features.

        Args:
            keep_only_sin_cos_transforms (bool): Whether to keep only the sin/cos transformation features and drop
                the plain date features.

        Returns:
            (self).
        """

        self.df['dayofweek_sin'] = self._sin_transform(self.df['dayofweek'])
        self.df['dayofweek_cos'] = self._cos_transform(self.df['dayofweek'])
        self.df['week_sin'] = self._sin_transform(self.df['week'])
        self.df['week_cos'] = self._cos_transform(self.df['week'])
        self.df['month_sin'] = self._sin_transform(self.df['month'])
        self.df['month_cos'] = self._cos_transform(self.df['month'])
        self.df['day_sin'] = self._sin_transform(self.df['day'])
        self.df['day_cos'] = self._cos_transform(self.df['day'])
        self.df['hour_sin'] = self._sin_transform(self.df['hour'])
        self.df['hour_cos'] = self._cos_transform(self.df['hour'])

        if keep_only_sin_cos_transforms:
            self.df.drop(columns=[
                'dayofweek', 'week', 'month', 'year', 'day', 'hour'
            ], inplace=True)

        return self

    def has_enough_records(self, days_in_hours):
        """
        Method that finds if the user's data have the required number of days (in hours) (after resampling) to be used
        later in the window.

        Args:
            days_in_hours (int): The days in hours.

        Returns:
            (bool): Whether user has enough records to be used later in the window.
        """

        start = self.df.index[0]
        end = self.df.index[-1]

        diff = end - start
        required_days = days_in_hours / 24

        if diff.days >= required_days:
            return True
        return False

    @staticmethod
    def remove_no_wear_days(df):
        """
        Removes days that user didn't wear the tracking device.
        Based on literature, no wear days are defined as steps less than 500.

        Essentially this function keeps all user's steps that are above 500 steps.

        Returns:
            (pd.DataFrame): The DataFrame with removed the no wear days.
        """

        df = df[df['var1(t)'] >= 500.0]
        return df

    @staticmethod
    def _sin_transform(values):
        """
        Applies SIN transform to a series value.

        Args:
            values (pd.Series): A series to apply SIN transform on.
        Returns
            (pd.Series): The transformed series.
        """

        return np.sin(2 * np.pi * values / len(set(values)))

    @staticmethod
    def _cos_transform(values):
        """
        Applies COS transform to a series value.

        Args:
            values (pd.Series): A series to apply SIN transform on.
        Returns
            (pd.Series): The transformed series.
        """

        return np.cos(2 * np.pi * values / len(set(values)))

    @staticmethod
    def __check_correctness(df):
        """
        Private method that ensures that the passed dataframe has the proper formatting, in order to be processed
        by the framework.

        Args:
            df (pd.DataFrame): The dataframe to perform checks.

        """

        assert 'value' in df.columns, "Dataframe must contain a column 'value' representing the data of " \
                                      "the time-series."

        if not is_datetime64_dtype(df.index.dtype):
            raise Exception("The index of the passed dataframe must be datetime data type.")

        if not (is_float_dtype(df.dtypes['value']) or is_int64_dtype(df.dtypes['value'])):
            raise Exception("The data type of 'value' column must be float or int.")



import pandas as pd


class Window:
    """
    Class the performs window operations.

    Attributes:
        n_in (int): The number of lagged observations for the window construction.
        n_out (int): The number of output columns. Defaults to 1 for uni-variate time series problems.
        dropna (bool): Whether to drop nan values due to non-existing previous data.
    """

    def __init__(self, n_in, n_out=1, dropna=True):
        self.n_in = n_in
        self.n_out = n_out
        self.dropna = dropna

    def sliding_window(self, data):
        """
        Function that turns a time series dataframe into a supervised dataset by using lagged observations.

        Args:
            data (pd.DataFrame): The data in a timeseries format. E.g. consecutive values.

        Returns:
            (pd.DataFrame): The supervised dataset. For uni-variate time series problems, the output column
                is named 'var1(t)'.
        """

        n_vars = 1 if type(data) is list else data.shape[1]

        new_cols = []
        new_names = []
        for i in range(self.n_in, 0, -1):
            lagged_obs = data.shift(i)
            new_cols.append(lagged_obs)

            for j in range(n_vars):
                name = "var{}(t-{})".format(j + 1, i)
                new_names.append(name)

        for i in range(0, self.n_out):
            new_cols.append(data.shift(-i))
            if i == 0:
                for j in range(n_vars):
                    name = ('var{}(t)'.format(j + 1))
                    new_names.append(name)
            else:
                for j in range(n_vars):
                    name = ('var{}(t+{})'.format(j + 1, i))
                    new_names.append(name)

        result = pd.DataFrame()

        for col in new_cols:
            result = pd.concat([result, col], axis=1)

        result.columns = new_names
        if self.dropna:
            result.dropna(inplace=True)

        return result

    def tumbling_window(self, data, freq='1D'):
        """
        Function that aggregates predictions for the next day by default.
        This is useful in a way that we want to predict the next day's steps, rather than
        the steps of the next hour (if resampling by hour).

        It return only the first from the n_in records, because the others are already included
        in the aggregated value. So sliding in the next day's values is cheating.

        Essentially it provides tumbling windows.

        Args:
            data (pd.DataFrame): The data in a timeseries format. E.g. consecutive values.
            freq (str): The frequency in panda's style of aggregation for resampling. Defaults to aggregation
                of next day's steps

        Returns:
            (pd.DataFrame): The dataset with output ('var1(t)' column) the aggregated steps count.
        """

        sliding_dataset = self.sliding_window(data=data)

        # find each days' total steps
        steps_per_day = sliding_dataset['var1(t)'].resample(freq).sum()

        # slide to replicate the tumbling window with no overlapping records
        tumbling_dataset = sliding_dataset[::24].copy()

        # drop hourly predictions
        tumbling_dataset.drop(columns=['var1(t)'], inplace=True)

        # create again the prediction column to be populated
        # with the aggregated values of the next day
        tumbling_dataset['var1(t)'] = None

        # populate target steps with the total aggregated steps of next day
        for date_index, steps in zip(tumbling_dataset.index, steps_per_day):
            tumbling_dataset.at[date_index, 'var1(t)'] = steps

        return tumbling_dataset



from sklearn.model_selection import train_test_split


class Dataset:

    """
    Class that provides sub-datasets for training ML models. It takes as input the dataset created from
    the UBIWEAR's pre-processor.

    Attributes:
        dataset (pd.DataFrame): The dataset created from Processor
    """

    def __init__(self, dataset):
        self.dataset = dataset

    def get_train_test(self, train_ratio=0.75):
        """
        Returns train and test data respecting the chronological order of the time series dataset.

        Args:
            train_ratio (float): The ratio for training/testing.

        Returns:
            (pd.DataFrame), (pd.DataFrame), (pd.DataFrame), (pd.DataFrame): The x_train, x_test, y_train,
                y_test sub-datasets.
        """

        y = self.dataset['var1(t)']
        X = self.dataset.drop(columns=['var1(t)'])

        # Split into train and test with respect to the chronological order i.e. no shuffle
        x_train, x_test, y_train, y_test = train_test_split(X, y, train_size=train_ratio, shuffle=False)

        return x_train, x_test, y_train, y_test

    def get_train_val_test(self, train_ratio=0.75, val_ratio=0.2):
        """
        Returns train, validation and test data respecting the chronological order of the time series dataset.

        It splits the initial training set into a new training and validation set.

        Args:
            train_ratio (float): The ratio for training/testing.
            val_ratio (float): The ratio for the validation set.

        Returns:
            (pd.DataFrame), (pd.DataFrame), (pd.DataFrame), (pd.DataFrame), (pd.DataFrame), (pd.DataFrame): The
            x_train, x_val, x_test, y_train, y_val, y_test sub-datasets.
        """

        x_train_val, x_test, y_train_val, y_test = self.get_train_test(train_ratio=train_ratio)

        x_train, x_val, y_train, y_val = train_test_split(x_train_val, y_train_val, test_size=val_ratio, shuffle=False)

        return x_train, x_val, x_test, y_train, y_val, y_test



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='ubiwear',
    version='0.0.16',
    url='https://github.com/stergiosbamp/deep-physical-activity-prediction/ubiwear/',
    description='A Python library for pre-processing ubiquitous aggregated self-tracking data',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license_files=['LICENSE'],
    author='Stergios Bampakis',
    author_email='bampakis.stergios@gmail.com',
    python_requires='>=3.8',
    classifiers=[
        'Programming Language :: Python :: 3.8',
        'Development Status :: 3 - Alpha',
        'Operating System :: OS Independent'
    ],
    install_requires=['pandas', 'scikit-learn'],
    packages=['ubiwear']
)
